#!/usr/bin/perl
##############################################################################
# Owner : Martin Rieder (martin.rieder@schwindt.eu)                          #
#                                                                            #
# Name               : sctliccheck.pl                                        #
# Version            : 3.00                                                  #
# Intention          : Read License Usage into XML                           #
# Require            : perl                                                  #
#                                                                            #
# v1.00 04.05.15 mrd : created                                               #
# v2.00 10.09.15 mrd : some enhancements                                     #
# v2.10 22.09.15 mrd : local licenses added                                  #
# v3.00 24.11.15 mrd : logging added and other improvements                  #
# v3.00 24.11.15 mrd : logging added and other improvements                  #
# v3.01 03.12.18 mrd : better handling of unc-paths                          #
##############################################################################

use strict;
use List::MoreUtils qw(firstidx uniq);
use XML::Simple;
use Getopt::Long;
use File::Spec::Functions;
use File::Basename;
use File::Glob ':glob';

my $RequiredPerlVersion=5.016;
my $script_path=dirname(__FILE__);
my $default_config=catfile($script_path, "..", "Config", "Licenses", "licenses.conf");
my $dslicsrv_command="$ENV{CATPRO_LOC}\\bin\\DSLS\\win_b64\\code\\bin\\DSLicSrv.exe";
my $dslicsrv_command_fallback="C:\\Program Files\\Dassault Systemes\\DS License Server\\win_b64\\code\\bin\\DSLicSrv.exe";
my $logfile="$ENV{CABUM_WORK}\\sctliccheck.log";
my $hide_non_startlic=1;
my $dsls_admin_port="4084";
my $dsls_config;
my $dslicmgtb_command;
my (@dsls_server_failover, @dsls_server_list, @lic, @target_id);
my (%opt, %count, %inuse, %usage, %expiration, %dsls_server, %exclude, %xml, %lic_is_slt);
my %months = ( 	Jan => '01',
				Feb => '02', 
				Mar => '03', 
				Apr => '04', 
				May => '05', 
				Jun => '06', 
				Jul => '07', 
				Aug => '08', 
				Sep => '09', 
				Oct => '10', 
				Nov => '11', 
				Dec => '12');

open(LOG, ">:encoding(UTF-8)", $logfile);

&logging("INFO", "sctliccheck.pl start logging...");

if ( $] < $RequiredPerlVersion ) {
	&logging("ERROR", "sctliccheck requires perl version $RequiredPerlVersion, found perl version $]")
} else {
	&logging("INFO", "found perl version $]")
}

if (-e $dslicsrv_command) {
	&logging("INFO", "license administration tool found (\"$dslicsrv_command\")");
} else {
	&logging("WARNING", "license administration tool not found (\"$dslicsrv_command\")");
	&logging("INFO", "searching for fallback (\"$dslicsrv_command_fallback\")");
	if (-e $dslicsrv_command_fallback) {
		&logging("INFO", "license administration tool found (\"$dslicsrv_command_fallback\")");
		$dslicsrv_command=$dslicsrv_command_fallback;
	} else {
		&logging("ERROR", "license administration tool not found (\"$dslicsrv_command_fallback\")");
		&logging("ERROR", "no server licenses available...");
	}
}

&GetOptions(\%opt, 'catinstallpath=s');

if ($opt{catinstallpath}) {
	&logging("INFO", "set catinstallpath to \"$opt{catinstallpath}\"");
} else {
	&logging("WARNING", "catinstallpath not set");
}

if ( defined($opt{catinstallpath}) ) {
	if ( -e (catfile($opt{catinstallpath}, "code", "bin", "DSLicMgtb.exe")) ) {
		$dslicmgtb_command=catfile($opt{catinstallpath}, "code", "bin", "DSLicMgtb.exe");
		&logging("INFO", "catinstallpath exists")
	} else {
		&logging("WARNING", "catinstallpath does not exist")
	}
	if ( -d (catfile($opt{catinstallpath}, "control")) ) {
		my $path=catfile($opt{catinstallpath}, "control");
		$path =~ s/\\/\//g;
		foreach (bsd_glob("$path/*.slt")) {
			$_ =~ /^.*\/(.*)\.slt$/;
			$lic_is_slt{$1}="1";
		}
		&logging("INFO", "list of start-licenses read");
	} else {
		&logging("WARNING", "list of start-licenses not read");
	}
}

if ( -e "$default_config" ) {
	&readconfig($default_config);
	&logging("INFO", "configuration file \"$default_config\" read");
} else {
	&logging("ERROR", "configuration file \"$default_config\" not found");
}

if ( -e "$ENV{CABUM_LICCONF}" ) {
	&readconfig($ENV{CABUM_LICCONF});
	&logging("INFO", "configuration file \"$ENV{CABUM_LICCONF}\" read");
} else {
	&logging("ERROR", "configuration file \"$ENV{CABUM_LICCONF}\" not found");
}

&logging("INFO", "setting hide_non_startlic=$hide_non_startlic");
&logging("INFO", "setting dsls_admin_port=$dsls_admin_port");

for (keys %exclude) {
	&logging("INFO", "exclude $_");
}

my $path=$ENV{CABUM_LICCONF_OUTPUT};
$path =~ s/\\/\//g;
foreach (glob("$path/cabum_licconf*.xml")) {
	&logging("INFO", "deleting $_...");
    unlink "$_";
}

if ($ENV{DSLS_CONFIG}) {
    $dsls_config=$ENV{DSLS_CONFIG};
} else {
    $dsls_config=$ENV{ALLUSERSPROFILE} . "\\DassaultSystemes\\Licenses\\DSLicSrv.txt";
}

if ($dsls_config) {
	&logging("INFO", "using \"$dsls_config\"...");
} else {
	&logging("WARNING", "no license server configuration found");
}

if (@dsls_server_failover) {
	&logging("INFO", "using license server failover from \"$ENV{CABUM_LICCONF}\"...");
	@dsls_server_failover=uniq(@dsls_server_failover);
    foreach (@dsls_server_failover) {
		my ($temp1, $temp2)=split(/:|,|>/, $_);
		push (@dsls_server_list, $temp1);
		$dsls_server{$temp1}=$temp2;
    }
} else {
    &get_license_server($dsls_config);
}

if (%dsls_server) {
	foreach (@dsls_server_list) {
		&logging("INFO", "license server $_ (port: $dsls_server{$_})");
	}
} else {
	&logging("WARNING", "no license servers defined");
}

foreach (@dsls_server_list) {
    &get_license_info($_, $dsls_server{$_});
}

if ($dslicmgtb_command) {&get_license_info_local};

&logging("INFO", "removing duplicate license entries...");
@lic=uniq(@lic);
&logging("INFO", "sorting license entries...");
@lic=sort(@lic);
&logging("INFO", scalar(@lic) . " licenses found");

&logging("INFO", "creating \"$ENV{CABUM_LICCONF_OUTPUT}\\cabum_licconf.xml\"");
foreach (@lic) {
    my $show="true";
	if ($hide_non_startlic == 1) {
		my $temp1 = $_;
		$temp1 =~ s/^(.{3}).*$/$1/;
		if (not $lic_is_slt{$temp1} == 1) {
			$show="false";
		}
	}
	if ($exclude{$_} == 1) {
	    $show="false";
	}
	if ($usage{$_}) {
		my @usage=split(/=/,$usage{$_});
		my %license_usage;
		for (my $i=0; $i<scalar(@usage); $i++) {
			my ($server,$computer,$user,$process,$since,$until)=split(/;/,$usage[$i]);
			my ($server_next,$computer_next,$user_next,$process_next,$since_next,$until_next)=split(/;/,$usage[$i+1]);
			if (($computer eq $computer_next) && ($user eq $user_next)) {
				$usage[$i]="$server;$computer;$user;$process*$process_next;$since*$since_next;$until";
				splice(@usage,$i+1,1);
				$i--;
			}
		}
		foreach (@usage) {
			my ($server,$computer,$user,$process,$since,$until)=split(/;/,$_);
			my (@process)=split(/\*/,$process);
			my (@since)=split(/\*/,$since);
		    push(@{$license_usage{'license_usage'}},{
				'Server' 	=> $server,
				'Computer' 	=> $computer,
				'User' 		=> $user,
				'Process'	=> \@process,
				#'Since' 	=> $since,
				'Since' 	=> \@since,
				'Until'		=> $until
			});
		}
	    push(@{$xml{'lic'}},{
	        'Name'		=> $_,
	        'Count'		=> $count{$_},
	        'In-Use'	=> $inuse{$_},
			'Show'		=> $show,
			%license_usage
		});
	} else {
		push(@{$xml{'lic'}},{
		    'Name'		=> $_,
		    'Count'		=> $count{$_},
		    'In-Use'	=> $inuse{$_},
			'Show'		=> $show,
		});
	}
}

&xml_output;

&logging("INFO", "...end logging");

close(LOG);

sub logging {
	my $temp1=$_[0];
	my $temp2=$_[1];
	chomp $temp1;
	chomp $temp2;
	print LOG localtime() . " - $temp1: $temp2\n";
}

sub readconfig {
	open(FILE,"$_[0]");
	my @config=<FILE>;
	close(FILE);
	for (my $i=0;$i<@config;$i++) {
		if ( ($config[$i] !~ m/^#+/) && ($config[$i] !~ /^\s*$/) ) { 
			if ( not rindex($config[$i],"=") == -1 ) {
				$config[$i] =~ s/^\s+//;
				$config[$i] =~ s/\s+$//;
				if ( substr($config[$i],0,rindex($config[$i],"=")) =~ m/license-server/i ) {
					push(@dsls_server_failover, substr($config[$i],rindex($config[$i],"=")+1,length($config[$i])-(rindex($config[$i],"=")+1)));
				}
				if ( substr($config[$i],0,rindex($config[$i],"=")) =~ m/admin-port/i ) {
					$dsls_admin_port = substr($config[$i],rindex($config[$i],"=")+1,length($config[$i])-(rindex($config[$i],"=")+1));
				}
				if ( substr($config[$i],0,rindex($config[$i],"=")) =~ m/exclude/i ) {
					push(my @exclude, split(",", substr($config[$i],rindex($config[$i],"=")+1,length($config[$i])-(rindex($config[$i],"=")+1))));
					foreach (@exclude) {
						$exclude{$_}=1;
					}
				}
				if ( substr($config[$i],0,rindex($config[$i],"=")) =~ m/hide_non_startlic/i ) {
					$hide_non_startlic = substr($config[$i],rindex($config[$i],"=")+1,1);
				}
			}
		}
	}
}

sub get_license_server {
    open(FILE, $_[0]);
    my @dsls=<FILE>;
    close(FILE);
	@dsls=uniq(@dsls);
    foreach (@dsls) {
		chomp($_);
		if ($_ !~ m/^#/) {
		    (my $temp)=split(/:|,|>/, $_);
			push(@dsls_server_list, $temp);
		    $dsls_server{$temp}=$dsls_admin_port;
		}
    }
}

sub get_license_info {
	&logging("INFO", "querying license server $_[0] (port: $_[1])...");
    #my @temp1=`\"$dslicsrv_command\" -admin -run \"c $_[0] $_[1] -r; glu -all; ea -o \"\"$ENV{CABUM_LICCONF_OUTPUT}\\cabum_licconf_rules_$_[0].xml\"\"; d\" -locale en_US`;
    my @temp1=`\"$dslicsrv_command\" -admin -run \"c $_[0] $_[1] -r; glu -all; d\" -locale en_US`;
	&logging("INFO", "license administration tool info: $temp1[0]");
	if ($temp1[3] =~ m/admin > glu -all/) {
		&logging("ERROR", $temp1[2]);
		return;
	}
	if ($temp1[2] =~ m/Enter password/) {
		&logging("ERROR", "remote access to server $_[0] is password protected");
		return;
	}
    my $tid=substr($temp1[6], index($temp1[6], "Server id")+11);
    $tid =~ s/\s//g;
	my $skip="0";
	foreach (@target_id) {
		if ($_ eq $tid) {
			$skip="1";
			&logging("INFO", "duplicate license server found! Skipping $_[0] ($tid)...");
		}
	}
	if ($skip=="0") {
		push (@target_id, $tid);
		if ($temp1[8] =~ m/admin > glu -all/) {
			&logging("INFO", "license server info:");
			&logging("INFO", $temp1[2]);
			&logging("INFO", $temp1[3]);
			&logging("INFO", $temp1[4]);
			&logging("INFO", $temp1[5]);
			&logging("INFO", $temp1[6]);
		}
		for (my $i = (firstidx {/^\tDassault Systemes V5/} @temp1)+1; $i < @temp1; $i++) {
			if ($temp1[$i] =~ m/^\t\t\w{3}/) {
				chomp $temp1[$i];
				my $trigram=substr($temp1[$i], 0, index($temp1[$i], "maxReleaseNumber"));
				$trigram =~ s/\s//g;
				my $count=substr($temp1[$i], index($temp1[$i], "count")+6, (index($temp1[$i], "inuse"))-(index($temp1[$i], "count")+6));
				$count =~ s/\s//g;
				my $inuse=substr($temp1[$i], index($temp1[$i], "inuse")+6, (index($temp1[$i], "customerId"))-(index($temp1[$i], "inuse")+6));
				$inuse =~ s/\s//g;
				push (@lic, "$trigram");
				$count{"$trigram"}=$count{"$trigram"} + $count;
				$inuse{"$trigram"}=$inuse{"$trigram"} + $inuse;
			}
			if ($temp1[$i] =~ m/^\t{4}.*$/) {
				my $until = "";
				chomp $temp1[$i];
				$temp1[$i] =~ s/^\t{4}//;
				my ($computer, $user);
				($computer,undef,$user)=split(/\s+/, $temp1[$i]);
				my $process = substr($temp1[$i], index($temp1[$i], $user)+length($user), (index($temp1[$i], "5."))-(index($temp1[$i], "$user")+length($user)));
				$process=~s/^\s*//;
				$process=~s/\s*$//;
				if ($process =~ /offline/) {$process="offline";}
				my $since = substr($temp1[$i], index($temp1[$i], "since:")+7, 18);
				print $since . "\n";
				# my($month, $day, $year, $time, $meridiem)=split(/ |\//, $since);
				# $day =~ s/,//g;
				# if ($day < 10) {$day="0" . $day};
				# #$month=$months{$month};
				# if ($month < 10) {$month="0" . $month};
				# $year="20" . $year;
				# my ($hours, $minutes) = split(/:/, $time);
				# if ($hours < 12) {$hours+=12};
				# if ($hours < 10) {$hours="0" . $hours};
				#$since="$year-$month-$day, $hours:$minutes";
				#$since="$day.$month.$year $hours:$minutes";
				#print $since . "\n";
				if ($process eq "offline") {
					$until = substr($temp1[$i], index($temp1[$i], "hold until:")+12, 14);
					#($month, $day, $year, $time, $meridiem)=split(/ |\//, $until);
					#$day =~ s/,//g;
					#if ($day < 10) {$day="0" . $day};
					#$month=$months{$month};
					#if ($month < 10) {$month="0" . $month};
					#$year="20" . $year;
					#my ($hours, $minutes) = split(/:/, $time);
					#if ($meridiem eq "PM" && $hours < 12) {$hours+=12};
					#if ($hours < 10) {$hours="0" . $hours};
					#$until="$year-$month-$day, $hours:$minutes";
				}
				$usage{$lic[-1]}=$usage{$lic[-1]} . "$_;$computer;$user;$process;$since;$until=";
			}
		}
	}
}

sub get_license_info_local {
	&logging("INFO", "querying local licenses...");
    my @temp1=`\"$dslicmgtb_command\"`;
    foreach(@temp1) {
	    chomp $_;
		if ($_) {
			my ($trigram,$server)=split(/ - /,$_);
			chomp $trigram;
			chomp $server;
			$trigram =~ s/\s//g;
			$server =~ s/\s//g;
			$trigram = $trigram . " (" . $server . ")";
			push (@lic, "$trigram");
		    $count{"$trigram"}="1";
		    $inuse{"$trigram"}="-";
		}
    }
}

sub xml_output {
	&logging("INFO", "writing \"$ENV{CABUM_LICCONF_OUTPUT}\\cabum_licconf.xml\"");
    open(FILE, ">:encoding(UTF-8)", "$ENV{CABUM_LICCONF_OUTPUT}\\cabum_licconf.xml");
    my $xsimple = XML::Simple->new(ForceArray=>1,rootname=>'license_config',noattr=>1,xmldecl=>'<?xml version="1.0" encoding="utf-8"?>');
    print FILE $xsimple->XMLout({
		'licenses'	=> [\%xml],
		'timestamp'	=> time
    });
    close(FILE);
}
